/***********************************************
 *  proc.c
 *
 *  Copyright 2001 Eric Kratzer.
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;
extern flag_t prompt_now, ignore_char_input;

/* main.c */
extern struct termios savtty, newtty;
extern int thisprgp;
extern flag_t exit_flag, piped_io_flag;

/************************************
 *        Global Variables          *
 ************************************/

sighandler_t sigchild_handler;

/************************************
 *        Local variables           *
 ************************************/

static char *JMODE[] = { 
    "Run (FG)",
    "Running",
    "Stopped",
    "(Internal player)",
    "(External player)",
    "(Pre loader)"
};
static char tmpstr_jb[JOBB_SIZE];
static DCharBuf *proc_list;

/************************************
 *         Module routines          *
 ************************************/

unsigned int proc_get_bytec()
{
    ProcessEntry *curr;
    unsigned int bytec;
    int i, len;

    proc_protect();
    bytec = 0;
    len = proc_get_n();

    for(i = 0; i < len; i++) {
	curr = proc_get_i(i);
	bytec += dbuf_get_bytec(curr->buff);
    }

    bytec += dbuf_get_bytec(proc_list);
    proc_unprotect();

    return bytec;
}

void proc_init()
{
    proc_list = dbuf_init(NULL, DCHAR_GROW, 32, 32);
    sigchild_handler = handle_child_sigs;
    memset(tmpstr_jb,  0, JOBB_SIZE);
}

void proc_clear()
{
    ProcessEntry *curr;
    int i, len;

    len = proc_get_n();

    for(i = 0; i < len; i++) {
	curr = proc_get_i(i);
	dbuf_free(curr->buff);
    }

    dbuf_free(proc_list);
}

/************************************
 *   Job Control functions          *
 ************************************/

void proc_protect()
{
    sigchild_handler = count_child_sigs;
    signal(SIGCHLD, (sighandler_t) sigchild_handler);
}

void proc_unprotect()
{
    sigchild_handler = handle_child_sigs;
    signal(SIGCHLD, (sighandler_t) sigchild_handler);
}

int proc_get_n()
{
    return proc_list->len;
}

ProcessEntry *proc_get_i(int index)
{
    return (ProcessEntry *) dbuf_get(proc_list, index);
}

ProcessEntry *proc_get(pid_t pid)
{
    ProcessEntry *curr;
    int i, len;

    len = proc_get_n();

    for(i = 0; i < len; i++) {
	curr = proc_get_i(i);
	if (curr->pid == pid)
	    return curr;
    }
    
    return NULL;
}

int proc_remove(int pid)
{
    ProcessEntry *curr;
    int i, len;

    len = proc_get_n();

    for(i = 0; i < len; i++) {
	curr = proc_get_i(i);
	if (curr->pid == pid) {
	    dbuf_free(curr->buff);
	    return dbuf_remove(proc_list, i);
	}
    }
    
    return FALSE;
}

int proc_stop_fg(ProcessEntry * curr)
{
    if (!curr)
	return -1;

    if (curr->mode < J_DECINT) {
	killpg(curr->pid, SIGSTOP);

	if (!piped_io_flag) {
	    tcsetpgrp(0, thisprgp);
	    tcgetattr(0, &curr->tty);
	    tcsetattr(0, TCSAFLUSH, &newtty);
	}

	curr->mode = J_STOPPED;
	ignore_char_input = FALSE;
	conb_add("\n" STATUS_MSG " %s [%d] %s", curr->name, curr->pid,
		 JMODE[curr->mode]);
    }

    return curr->mode;
}

void proc_set_term(ProcessEntry * curr)
{
    switch (curr->mode) {
    case J_FGRUN:
	ignore_char_input = FALSE;

	if (!piped_io_flag) {
	    tcsetpgrp(0, thisprgp);
	    tcsetattr(0, TCSAFLUSH, &newtty);
	}
	break;
    case J_BGRUN:
    case J_STOPPED:
	conb_add(STATUS_MSG " %s [%d] done", curr->name, curr->pid);
	break;
    }
}

int proc_clean(ProcessEntry * curr)
{
    int jobt;

    jobt = -1;

    if (!curr)
	int_sig_handler();
    else {
	proc_clean_pipes(curr);
	proc_set_term(curr);
	jobt = curr->mode;
	proc_remove(curr->pid);
    }

    return jobt;
}

int proc_get_stream(int fd, ProcessEntry * pr)
{
    int len;

    len = read(fd, tmpstr_jb, (JOBB_SIZE - 1));

    if (len > 0) {
	tmpstr_jb[len] = ZCHAR;
	if (pr->mode == J_FGRUN)
	    out_str_n(tmpstr_jb);
	else
	    dbuf_add_str(pr->buff, tmpstr_jb);
    }

    return len;
}

void proc_check_pipes(fd_set * read_set)
{
    ProcessEntry *curr;
    int i, len;

    proc_protect();
    len = proc_get_n();

    for(i = 0; i < len; i++) {

	curr = proc_get_i(i);

	if (curr->redir) {
	    if (FD_ISSET(curr->pipestdout[1], read_set))
		proc_get_stream(curr->pipestdout[0], curr);

	    if (FD_ISSET(curr->pipestderr[1], read_set))
		proc_get_stream(curr->pipestderr[0], curr);

	}
    }

    proc_unprotect();
}

void proc_clean_pipes(ProcessEntry * pr)
{
    DCharBuf *buf;
    struct timeval tv;
    fd_set read_set;
    int i, len, len2, ret, maxfd;;

    if (!pr || !pr->redir)
	return;
	
    do {
	do {
	    FD_ZERO(&read_set);
	    len = len2 = maxfd = 0;

	    FD_SET(pr->pipestdout[1], &read_set);
	    maxfd = pr->pipestdout[1] + 1;

	    FD_SET(pr->pipestderr[1], &read_set);
	    if (maxfd <= pr->pipestderr[1])
		maxfd = pr->pipestderr[1] + 1;

	    tv.tv_sec = tv.tv_usec = 0;

	    ret = select(maxfd, &read_set, NULL, NULL, &tv);

	}
	while (ret == -1 && errno == EINTR);

	if (FD_ISSET(pr->pipestdout[1], &read_set))
	    len = proc_get_stream(pr->pipestdout[0], pr);
	
	if (FD_ISSET(pr->pipestderr[1], &read_set))
	    len2 = proc_get_stream(pr->pipestderr[0], pr);
	
    }
    while (len > 0 || len2 > 0);

    /* don't flush external players' buffers?? anybody? */
    if (pr->buff->len && pr->mode < J_DECINT) {
	buf = pr->buff;
	len = pr->buff->len;

	conb_add(STATUS_MSG " Buffered output from \"%s\" [pid=%d]:", 
		 pr->name, pr->pid);

	for (i = 0; i < len; i++)
	    conb_add_n(dbuf_get(buf, i));
    }

    close(pr->pipestdout[0]);
    close(pr->pipestderr[0]);
    close(pr->pipestdout[1]);
    close(pr->pipestderr[1]);
}

void proc_set_pipes(fd_set * read_set, int *maxfd)
{
    ProcessEntry *curr;
    int i, len;

    proc_protect();
    len = proc_get_n();

    for(i = 0; i < len; i++) {

	curr = proc_get_i(i);
	
	if (curr->redir) {
	    FD_SET(curr->pipestdout[1], read_set);
	    if (*maxfd <= curr->pipestdout[1])
		*maxfd = curr->pipestdout[1] + 1;
	    FD_SET(curr->pipestderr[1], read_set);
	    if (*maxfd <= curr->pipestderr[1])
		*maxfd = curr->pipestderr[1] + 1;
	}
    }

    proc_unprotect();
}

void proc_add(pid_t pid, int mode, int redir,
	      int stdo[2], int stdr[2], char *name)
{
    ProcessEntry curr;

    proc_protect();

    memset(&curr, 0, sizeof(ProcessEntry));

    curr.redir = redir;
    curr.pid = pid;
    curr.mode = mode;
    safe_strcpy(curr.name, name, 32);

    if (redir) {
	curr.buff = dbuf_init(NULL, DCHAR_SHIFT, 256, 0);
	curr.pipestdout[0] = stdo[0];
	curr.pipestdout[1] = stdo[1];
	curr.pipestderr[0] = stdr[0];
	curr.pipestderr[1] = stdr[1];
    } else {
	curr.buff = NULL;
	curr.pipestdout[0] = -1;
	curr.pipestdout[1] = -1;
	curr.pipestderr[0] = -1;
	curr.pipestderr[1] = -1;
    }

    dbuf_add(proc_list, (void *) &curr, sizeof(ProcessEntry), 0);

    switch (mode) {
    case J_FGRUN:
	ignore_char_input = TRUE;
	if (!piped_io_flag)
	    tcsetattr(0, TCSAFLUSH, &savtty);
	break;
    case J_BGRUN:
	conb_add(STATUS_MSG " %s [%d] &", name, pid);
	break;
    }

    proc_unprotect();
}

void proc_flush(ProcessEntry * pr)
{
    DCBNode *buf;
    int i, len;

    if (pr->buff) {
	
	len = pr->buff->len;
	buf = pr->buff->buf;

	for (i = 0; i < len; i++)
	    out_str_n(buf[i].ptr);

	dbuf_clear(pr->buff);
    }
}

void proc_set_xg(ProcessEntry * curr, int mode)
{
    if (mode == J_FGRUN) {
	if (curr->redir)
	    proc_flush(curr);

	if (!piped_io_flag) {
	    tcsetpgrp(0, curr->pid);
	    if (curr->mode == J_STOPPED)
		tcsetattr(0, TCSAFLUSH, &curr->tty);
	    else
		tcsetattr(0, TCSAFLUSH, &savtty);
	}

	prompt_now = FALSE;
	ignore_char_input = TRUE;
    } else {
	ignore_char_input = FALSE;

	if (!piped_io_flag) {
	    tcsetpgrp(0, thisprgp);
	    tcsetattr(0, TCSAFLUSH, &newtty);
	}

	show_user(STATUS_MSG " %s [%d] &", curr->name, curr->pid);
    }

    curr->mode = mode;
    killpg(curr->pid, SIGCONT);
}

void proc_send_2xg(int mode)
{
    ProcessEntry *curr;
    char *num;
    int i, len, index;

    num = dbuf_get(args, 1);
    len = proc_get_n();

    if (!len) {
	show_user(DEBUG_MSG " Process list is empty");
	return;
    } else {
	
	if (num) {
	    index = atoi(num);
	    if (index <= 0)
		show_user(ERROR_MSG " Invalid <job_number>");
	    else {
		curr = proc_get_i(index - 1);
		if (!curr || curr->mode >= J_DECINT)
		    show_user(DEBUG_MSG " No viable process found");
		else
		    proc_set_xg(curr, mode);
	    }
	} else {
	    for(i = 0; i < len; i++) {
		curr = proc_get_i(i);
		if (curr->mode < J_DECINT) {
		    proc_set_xg(curr, mode);
		    return;
		}
	    }
	    show_user(DEBUG_MSG " No viable process found");
	}
    }
}

void /* user */ proc_2fg_comm()
{
    proc_protect();
    proc_send_2xg(J_FGRUN);
    proc_unprotect();
}

void /* user */ proc_2bg_comm()
{
    proc_protect();
    proc_send_2xg(J_BGRUN);
    proc_unprotect();
}

void /* user */ proc_flush_comm()
{
    ProcessEntry *curr;
    char *ndx;
    int num;

    ndx = dbuf_get(args, 1);

    if (!ndx || !is_all_digits(ndx)) {
	print_cmnd_help(FLUSH_JOB_BUFF);
	return;
    }

    proc_protect();
    num = atoi(ndx);

    curr = proc_get_i(num - 1);

    if (!curr)
	show_user(ERROR_MSG " Invalid job number: %d", num);
    else if (!curr->buff)
	show_user(DEBUG_MSG " No Buffer allocated for \"%s\"", curr->name);
    else if (curr->buff->len == 0)
	show_user(DEBUG_MSG " Buffer for \"%s\" is empty", curr->name);
    else {
	show_user(STATUS_MSG " Flushing job buffer for \"%s\" [pid=%d]:",
		  curr->name, curr->pid);

	proc_flush(curr);
    }

    proc_unprotect();
}

void /* user */ proc_print_comm()
{
    ProcessEntry *curr;
    char *tmp;
    int i, len;

    proc_protect();
    len = proc_get_n();

    if (!len)
	outb_add(DEBUG_MSG " Process List is empty");
    else {
	for(i = 0; i < len; i++) {

	    curr = proc_get_i(i);

	    outb_add(MISC_MSG " [%d] \"%s\"", i+1, curr->name);
	    outb_add("\tPID     : %d", curr->pid);
	    outb_add("\tState   : %s", JMODE[curr->mode]);
	    if (curr->buff) {
		if (curr->buff->len)
		    tmp = "[*]";
		else
		    tmp = "[ ]";
		outb_add("\tBuffer  : %s (%d bytes allocated)", tmp,
			 curr->buff->bytec);
	    } else
		outb_add("\tBuffer  : None");
	}
    }

    proc_unprotect();
    outb_flush();
}
